"""Util class for building DFCX webhook payloads."""

# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import string

# logging config
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(levelname)-8s %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S')

class WebhookUtils():
    """Utils class for quickly building DFCX webhook payloads."""

    @staticmethod
    def build_page_info(param_name, param_state):
        """Builds a Page Info object for Dialogflow CX.

        JSON format expected by Dialogflow CX as a Response payload.

        Args:
          param_name: The name of the Dialogflow CX parameter to update
          param_state: One of the following states [EMPTY, INVALID, FILLED]

        Returns:
          page_info: A JSON object in the appropriate Response format
        """
        page_info = {
            'formInfo': {
                'parameterInfo': [
                    {
                        'displayName': param_name,
                        'state': param_state
                    }
                ]
            }
        }

        return page_info

    @staticmethod
    def build_session_info(parameters):
        """Builds a Session Info object for Dialogflow CX Response.

        Provides the ability to perform CRUD functions for the DFCX Session
        parameters by building a JSON object to be included in the final build
        Response sent back to DFCX.

        Args:
          paramters: JSON formatted key:value pairs to be updated in the
            session params

        Returns:
          session_info: JSON object formatted for the DFCX Fulfillment Response
            Body

        Example Input: {'key1':'value1', 'key2':'value2'}
        """
        session_info = {'parameters': parameters}

        return session_info

    @staticmethod
    def build_response(response_text=None, page_info=None, session_info=None):
        """Builds a Response object for Dialogflow CX.

        Provides the JSON object structure expected by DFCX for the Response
        payload, including optional paload for pageInfo and sessionInfo
        depending on the use case.

        Args:
          response_text: The text response to be displayed to the user. Can
            also be empty string if no response to the user is required.
          page_info: (Optional) The JSON object returned by build_page_info()
          session_info: (Optiona) The JSON object returned by
            build_session_info()
        """
        if response_text:
            response_object = {
                'mergeBehavior': 'REPLACE',
                'messages': [
                    {
                        'text': {
                            'text': [response_text]
                        }
                    }
                ]
            }

        else:
            response_object = None

        message = {
            'fulfillmentResponse': response_object,
            'pageInfo': page_info,
            'sessionInfo': session_info
        }

        return message

    @staticmethod
    def get_tag(request):
        """Get the incoming DFCX webhook tag."""

        return request['fulfillmentInfo']['tag']

    @staticmethod
    def get_parameters(request):
        """Get the incoming DFCX webhook params."""
        params = None

        if 'parameters' in request['sessionInfo']:
            params = request['sessionInfo']['parameters']

        return params

    @staticmethod
    def get_conf_score(request):
        """Get the intent confidence score of the last triggered intent."""

        return request['intentInfo']['confidence']

    @staticmethod
    def get_user_utterance(request, cleaned=False):
        """Get the user utterance delivered with the webhook Request.

        Args:
          request: The incoming DFCX Request
          cleaned: Boolean flag with the following functions
            False - The user utterance is returned verbatim.
            True - The user utterance is stripped of whitespace,
              punctuation, and returned as lowercase.

        Returns:
          user_utterance: The final user utterance that was captured.
        """

        if cleaned:
            user_utterance = request['text'].lower().translate(str.maketrans(
                '', '', string.punctuation))

        else:
            user_utterance = request['text']

        return user_utterance
        


"""A sample script for building a DFCX bot from scratch using simple inputs."""
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import pandas as pd

from dfcx_scrapi.core.agents import Agents
from dfcx_scrapi.core.intents import Intents
from dfcx_scrapi.core.flows import Flows
from dfcx_scrapi.core.pages import Pages
from dfcx_scrapi.tools.dataframe_functions import DataframeFunctions
from dfcx_scrapi.tools.maker_util import MakerUtil

def build_agent(creds_path, project_id):
    """Build a simple agent."""

    # First we will instansiate our Agent object
    a = Agents(creds_path=creds_path)

    # Next, we will set some variables for our agent creation args
    display_name = 'My Cool Agent!'
    gcp_region = 'us-central1'

    # Then we will call the `create_agent` method and capture the result in a
    # var called `my_agent`
    my_agent = a.create_agent(project_id, display_name, gcp_region)

    # First, we will instantiate our DataframeFunctions (dffx) object
    dffx = DataframeFunctions(creds_path=creds_path)

    # Next, we will read in our sample CSV with Intent/TP data into a Pandas
    # DataFrame
    df = pd.read_csv('../data/sample_intent_tp.csv')

    # Finally, we will use `dffx` to push our Intents to our Agent
    dffx.bulk_create_intent_from_dataframe(
        my_agent.name, df, update_flag=True)

    # First, we will instantiate our Flows and Pages objects.
    f = Flows(creds_path=creds_path)
    p = Pages(creds_path=creds_path)

    # The `get_flows_map` method provides an easy to use map of your Flows,
    # their IDs, and their human readable Display Names
    # Using the `reverse=True` arg allows you to call the dictionary by your
    # flow Display Names which can be easier for exploratory building.
    flows_map = f.get_flows_map(my_agent.name, reverse=True)

    # Now that we know our Flow ID, we'll create a Page using the Flow ID as
    # the Page parent
    my_page = p.create_page(
        flows_map['Default Start Flow'],
        display_name='My First Page!')

    # We'll first get the Default Start Flow (dsf for short)
    dsf = f.get_flow(flows_map['Default Start Flow'])

    # We'll also fetch our Intent directly from the bot by Display Name to make
    # this next part easier
    i = Intents(creds_path=creds_path)
    intent_list = i.list_intents(my_agent.name)
    my_intent = [
        x for x in intent_list if \
            x.display_name == 'head_intent.order_pizza'][0]

    # Next, we'll use the MakerUtil to build our Transition Route (tr) object
    mu = MakerUtil()
    my_tr = mu.make_transition_route(
        intent=my_intent.name, target_page=my_page.name)

    # Now, we'll add the newly created Transition Route object to our DSF
    # object
    dsf.transition_routes.append(my_tr)

    # Finally, we'll update the DSF object in our agent
    f.update_flow(
        flows_map['Default Start Flow'],
        obj=dsf,
        transition_routes=dsf.transition_routes)

if __name__ == '__main__':
    CREDS_PATH = str(sys.argv[1])
    PROJECT_ID = str(sys.argv[2])
    build_agent(CREDS_PATH, PROJECT_ID)



"""Sample Cloud Function code to extract and return DFCX Confidence Score."""

# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
from dfcx_scrapi.tools.webhook_utils import WebhookUtils

# logging config
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(levelname)-8s %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S')


def conf_score(request, debugging=False):
    """Extract the Confidence Score and return as a Session Parameter.

    This Cloud Function code is meant to work in conjunction with a Dialogflow
    CX agent. The `conf_score` function is the main entry point and receives
    the incoming webhook request from Dialogflow CX. The request is then parsed
    and
    """
    logging.info('###### [cx_webhook] - CFX Triggered! ######')
    wu = WebhookUtils()

    if debugging:
        req = request

    else:
        req = request.get_json()
        logging.info('* [cx_webhook] Incoming Request: %s', req)

    webhook_tag = wu.get_tag(req)

    if webhook_tag == 'conf_test':
        logging.info('* [cx_webhook] CONF TEST TAG RECEIVED.')
        score = wu.get_conf_score(req)
        session_info = wu.build_session_info({'webhook_conf': score})
        message = wu.build_response('The Confidence Score is: {}'.format(
          score), session_info=session_info)

    else:
        logging.info('* [cx_webhook] No Webhook Tag Received.')

    res = json.dumps(message)
    logging.info(res)

    return res

if __name__ == '__main__':
    FILE = '/Users/pmarlow/eng/cloud_functions/tests/conf_score_payload.json'
    with open(FILE) as json_file:
        data = json.load(json_file)

    conf_score(data, False)



"""Flow Resource functions."""

# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict
import requests
import google.cloud.dialogflowcx_v3beta1.services as services
import google.cloud.dialogflowcx_v3beta1.types as types
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core.scrapi_base import ScrapiBase

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Flows(ScrapiBase):
    """Core Class for CX Flow Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        flow_id: str = None,
        agent_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        if flow_id:
            self.flow_id = flow_id

        self.agent_id = agent_id

    def get_flows_map(self, agent_id: str = None, reverse=False):
        """Exports Agent Flow Names and UUIDs into a user friendly dict.

        Args:
            - agent_id, the formatted CX Agent ID to use
            - reverse, (Optional) Boolean flag to swap key:value -> value:key

        Returns:
            - flows_dict, Dictionary containing flow UUIDs as keys and
                flow.display_name as values
        """
        if not agent_id:
            agent_id = self.agent_id

        if reverse:
            flows_dict = {
                flow.display_name: flow.name
                for flow in self.list_flows(agent_id=agent_id)
            }

        else:
            flows_dict = {
                flow.name: flow.display_name
                for flow in self.list_flows(agent_id=agent_id)
            }

        return flows_dict

    def train_flow(self, flow_id: str = None):
        """trains the specified flow.

        Args:
          flow_id: CX flow ID string in the following format
            projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>/
              flows/<FLOW ID>

        Returns:
          response: A Long Running Operation (LRO) ID that can be used to
            check the status of the export using
              dfcx_scrapi.core.operations->get_lro()
        """
        if not flow_id:
            flow_id = self.flow_id

        request = types.flow.TrainFlowRequest()
        request.name = flow_id

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options)

        response = client.train_flow(request)

        return response

    def list_flows(self, agent_id: str = None):
        """Get a List of all Flows in the current Agent.

        Args:
          agent_id, CX Agent ID string in the proper format
            projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>

        Returns:
          flows, a List of Flow objects
        """
        if not agent_id:
            agent_id = self.agent_id

        request = types.flow.ListFlowsRequest()
        request.parent = agent_id

        client_options = self._set_region(agent_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.list_flows(request)

        flows = []
        for page in response.pages:
            for flow in page.flows:
                flows.append(flow)

        return flows

    def get_flow(self, flow_id: str = None):
        """Get a single CX Flow object.

        Args:
          flow_id, CX Flow ID in the proper format

        Returns:
          response, a single CX Flow object"""
        if not flow_id:
            flow_id = self.flow_id

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.get_flow(name=flow_id)

        return response

    def update_flow(
        self, flow_id: str = None, obj: types.Flow = None, **kwargs
    ):
        """Update a single specific CX Flow object.

        Args:
          flow_id, CX Flow ID in the proper format
          obj, (Optional) a single CX Flow object of types.Flow

        Returns:
          response, a copy of the updated Flow object
        """

        if obj:
            flow = obj
            flow.name = flow_id
        else:
            if not flow_id:
                flow_id = self.flow_id
            flow = self.get_flow(flow_id)

        # set flow attributes to args
        for key, value in kwargs.items():
            setattr(flow, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.update_flow(flow=flow, update_mask=mask)

        return response

    def update_nlu_settings(self, flow_id, **kwargs):
        """updates flow to new NLU setting.
        Args:
            flow_id: flow id to update nlu settings for.
            model_type: (Optional) [0:unspecified, 1:MODEL_TYPE_STANDARD,
              2:Custom, 3:Advanced]
            classification_threshold: (Optional) threshold for the flow
            model_training_mode: (Optional) [0:unspecified, 1:automatic,
              2:'manual]
        """

        flow = self.get_flow(flow_id)
        current_settings = flow.nlu_settings
        for key, value in kwargs.items():
            setattr(current_settings, key, value)
        self.update_flow(flow_id=flow_id, nlu_settings=current_settings)

    def export_flow(
        self,
        flow_id: str,
        gcs_path: str,
        data_format: str = "BLOB",
        ref_flows: bool = True,
    ) -> Dict[str, str]:
        """Exports DFCX Flow(s) into GCS bucket.

        Args:
          flow_id, the formatted CX Flow ID to export
          gcs_path, the full GCS Bucket and File name path
          data_format, (Optional) One of 'BLOB' or 'JSON'. Defaults to 'BLOB'.
          ref_flows, (Optional) Bool to include referenced flows connected to
            primary flow

        Returns:
          lro, Dict with value containing a Long Running Operation UUID that
            can be used to retrieve status of LRO from
              dfcx_scrapi.core.operations->get_lro()
        """

        location = flow_id.split("/")[3]
        if location != "global":
            base_url = "https://{}-dialogflow.googleapis.com/v3beta1".format(
                location
            )
        else:
            base_url = "https://dialogflow.googleapis.com/v3beta1"
        url = "{0}/{1}:export".format(base_url, flow_id)

        body = {
            "flow_uri": "{}".format(gcs_path),
            "data_format": data_format,
            "include_referenced_flows": ref_flows,
        }

        headers = {
            "Authorization": "Bearer {}".format(self.token),
            "Content-Type": "application/json; charset=utf-8",
        }

        # Make REST call
        response = requests.post(url, json=body, headers=headers)
        response.raise_for_status()

        lro = response.json()

        return lro

    def import_flow(
        self,
        destination_agent_id: str,
        gcs_path: str,
        import_option: str = "FALLBACK",
    ) -> Dict[str, str]:
        """Imports a DFCX Flow from GCS bucket to CX Agent.

        Args:
          agent_id, the DFCX formatted Agent ID
          gcs_path, the full GCS Bucket and File name path
          import_option, one of 'FALLBACK' or 'KEEP'. Defaults to 'FALLBACK'

        Returns:
          lro, Dict with value containing a Long Running Operation UUID that
            can be used to retrieve status of LRO from
              dfcx_scrapi.core.operations->get_lro()
        """

        location = destination_agent_id.split("/")[3]
        if location != "global":
            base_url = "https://{}-dialogflow.googleapis.com/v3beta1".format(
                location
            )
        else:
            base_url = "https://dialogflow.googleapis.com/v3beta1"
        url = "{0}/{1}/flows:import".format(base_url, destination_agent_id)

        body = {
            "flow_uri": "{}".format(gcs_path),
            "import_option": "{}".format(import_option),
        }

        headers = {
            "Authorization": "Bearer {}".format(self.token),
            "Content-Type": "application/json; charset=utf-8",
        }

        # Make REST call
        response = requests.post(url, json=body, headers=headers)
        response.raise_for_status()

        lro = response.json()

        return lro

    def delete_flow(self, flow_id: str = None, force: bool = False) -> None:
        """Deletes a single CX Flow Object resources.

        Args:
          flow_id: flow to delete
          force: False means a flow will not be deleted if a route to the flow
            exists, True means the flow will be deleted and all
        """
        if not flow_id:
            flow_id = self.flow_id

        request = types.DeleteFlowRequest()
        request.name = flow_id
        request.force = force

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )

        client.delete_flow(request)



"""util class for doing searches"""

# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from typing import Dict

import pandas as pd

from dfcx_scrapi.core.entity_types import EntityTypes
from dfcx_scrapi.core.flows import Flows
from dfcx_scrapi.core.intents import Intents
from dfcx_scrapi.core.pages import Pages
from dfcx_scrapi.core.scrapi_base import ScrapiBase

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class SearchUtil(ScrapiBase):
    """class for searching items"""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        agent_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.intents = Intents(creds_path=creds_path, creds_dict=creds_dict)
        self.entities = EntityTypes(
            creds_path=creds_path, creds_dict=creds_dict
        )
        self.flows = Flows(creds_path=creds_path, creds_dict=creds_dict)
        self.pages = Pages(creds_path=creds_path, creds_dict=creds_dict)

        if agent_id:
            self.agent_id = agent_id
            self.flow_map = self.flows.get_flows_map(
                agent_id=agent_id, reverse=True
            )
            self.client_options = self._set_region(agent_id)

    def _find_true_routes_flow_level(self, flow_display_name, flow_map):
        flow_id = flow_map[flow_display_name]
        start_page = self.flows.get_flow(flow_id) # pylint: disable=W0612
        other_pages = self.pages.list_pages(flow_id)

        # Start page - no entry fulfillment
        pages_dataframe = pd.DataFrame()
        for page in other_pages:
            display_name = page.display_name

            webhook = False
            if page.entry_fulfillment.webhook:
                webhook = True

            has_parameters = False
            if page.form.parameters:
                has_parameters = True

            has_true_route = False
            has_true_final_route = False
            for route in page.transition_routes:
                if route.condition == "true":
                    has_true_route = True

                if route.condition == '$page.params.status = "FINAL" AND true':
                    has_true_final_route = True

            page_dataframe = pd.DataFrame(
                columns=[
                    "flow_display_name",
                    "page_display_name",
                    "webhook_entry_fullfillment",
                    "has_parameters",
                    "has_true_route",
                    "has_true_and_final_route",
                ],
                data=[
                    [
                        flow_display_name,
                        display_name,
                        webhook,
                        has_parameters,
                        has_true_route,
                        has_true_final_route,
                    ]
                ],
            )
            pages_dataframe = pages_dataframe.append(page_dataframe)

        return pages_dataframe

    # Flows - event handlers
    def _flow_level_handlers(self):
        flows_in_agent = self.flows.list_flows(self.agent_id)

        flow_event_handler_data = pd.DataFrame()
        for flow in flows_in_agent:
            flow_level_event_handlers = flow.event_handlers
            flow_level_event_handlers_dataframe = pd.DataFrame()

            for handler in flow_level_event_handlers:
                flow_level_event_handlers_dataframe = (
                    flow_level_event_handlers_dataframe.append(
                        pd.DataFrame(
                            columns=[
                                "flow",
                                "event",
                                "messages",
                                "transition_flow",
                                "transition_page",
                            ],
                            data=[
                                [
                                    flow.display_name,
                                    handler.event,
                                    handler.trigger_fulfillment.messages,
                                    handler.target_flow,
                                    handler.target_page,
                                ]
                            ],
                        )
                    )
                )
                flow_event_handler_data = flow_event_handler_data.append(
                    flow_level_event_handlers_dataframe
                )

        return flow_event_handler_data

    # Pages - event handlers
    def _page_level_handlers(self):
        page_level_event_handlers_all_dataframe = pd.DataFrame()
        flow_map = self.flows.get_flows_map(self.agent_id)
        for flow_ in flow_map.keys():
            pages_in_flow = self.pages.list_pages(flow_)

            for page in pages_in_flow:
                page_level_event_handlers = page.event_handlers
                page_level_event_handlers_dataframe = pd.DataFrame()
                for handler in page_level_event_handlers:
                    page_level_event_handlers_dataframe = (
                        page_level_event_handlers_dataframe.append(
                            pd.DataFrame(
                                columns=[
                                    "flow",
                                    "page",
                                    "event",
                                    "messages",
                                    "transition_flow",
                                    "transition_page",
                                ],
                                data=[
                                    [
                                        flow_map[flow_],
                                        page.display_name,
                                        handler.event,
                                        handler.trigger_fulfillment.messages,
                                        handler.target_flow,
                                        handler.target_page,
                                    ]
                                ],
                            )
                        )
                    )

                page_level_event_handlers_all_dataframe = (
                    page_level_event_handlers_all_dataframe.append(
                        page_level_event_handlers_dataframe
                    )
                )
        return page_level_event_handlers_all_dataframe

    # Parameters - event handlers
    def _parameter_level_handlers(self):
        parameter_level_event_handlers_all_dataframe = pd.DataFrame()
        flow_map = self.flows.get_flows_map(self.agent_id)
        for flow_ in flow_map.keys():
            pages_in_flow = self.pages.list_pages(flow_)
            for page in pages_in_flow:
                parameters = page.form.parameters
                for parameter in parameters:
                    parameter_event_handlers = (
                        parameter.fill_behavior.reprompt_event_handlers
                    )
                    param_lvl_event_df = pd.DataFrame()
                    for handler in parameter_event_handlers:
                        param_lvl_event_df = param_lvl_event_df.append(
                            pd.DataFrame(
                                columns=[
                                    "flow",
                                    "page",
                                    "parameter",
                                    "event",
                                    "messages",
                                    "transition_flow",
                                    "transition_page",
                                ],
                                data=[
                                    [
                                        flow_map[flow_],
                                        page.display_name,
                                        parameter.display_name,
                                        handler.event,
                                        handler.trigger_fulfillment.messages,
                                        handler.target_flow,
                                        handler.target_page,
                                    ]
                                ],
                            )
                        )
                    parameter_level_event_handlers_all_dataframe = (
                        parameter_level_event_handlers_all_dataframe.append(
                            param_lvl_event_df
                        )
                    )
        return parameter_level_event_handlers_all_dataframe


    def find_list_parameters(self, agent_id):
        """This method extracts Parameters set at a page level that are
        designated as "lists".

        Page level parameters are tied to Entity Types and can be returned
        as String or List types. If the user selects "list" at the page
        level, the Entity Type will be returned with "is_list: True". This
        function will allow the user to provide an Agent ID and will return
        all instances of parameters being used as lists on pages.

        Args:
          - agent_id, the Agent ID string in the following format:
            projects/<project_id>/locations/<location_id>/agents/<agent_id>

        Returns:
          - params_map, a Dict of parameter names and Pages they belong to
        """

        # entities = self.dfcx.list_entity_types(agent_id)
        flows_map = self.flows.get_flows_map(agent_id)

        params_list = []

        for flow in flows_map.keys():
            temp_pages = self.pages.list_pages(flow)
            for page in temp_pages:
                for param in page.form.parameters:
                    if param.is_list:
                        params_list.append(param.display_name)

        return params_list

    def search_conditionals_page(self, page_id, search):
        """search page for an exact string in conditional routes

        Args:
          - page_id, the formatted CX Page ID to use
          - search, string to search

        Returns:
          - locator, dataframe of the results of where this string was found
        """

        locator = pd.DataFrame()
        page = self.pages.get_page(page_id=page_id)
        i = 1
        for route in page.transition_routes:
            if search.lower() in route.condition.lower():
                iter_frame = pd.DataFrame(
                    columns=["resource_id", "condition", "route_id"],
                    data=[[page_id, route.condition, i]],
                )
                locator = locator.append(iter_frame)
            i += 1

        return locator

    def search_conditionals_flow(self, flow_id, search):
        """search flow for an exact string in conditional routes

        Args:
        - flow_id, the formatted CX Flow ID to use
        - search, string to search

        Returns:
        - locator, dataframe of the results of where this string was found
        """

        locator = pd.DataFrame()
        flow = self.flows.get_flow(flow_id=flow_id)
        i = 1
        for route in flow.transition_routes:
            if search.lower() in route.condition.lower():
                iter_frame = pd.DataFrame(
                    columns=["resource_id", "condition", "route_id"],
                    data=[[flow_id, route.condition, i]],
                )
                locator = locator.append(iter_frame)
            i += 1

        return locator

    def search_conditionals(
        self,
        search,
        agent_id,
        flow_name=None,
        page_name=None,
        flag_search_all=False,
    ):
        """This is the master function where a user can search across
        all pages in a flow, an entire agent etc.
        Search conditionals for an exact string in conditional routes.

        Args:
            - search, string to search
            - agent_id, the formatted CX Agent ID to use
            - flow_name, (optional) the display name of the flow to search
            - page_name,  (optional) the display name of the page to search
            - flag_search_all, (optional)
                When set to True:
                    -if just an agent_id then entire agent is searched
                    -if just an agent_id and flow_name are specified
                        an entire flow is searched
                    -if an agent_id, flow_name and page_name are specified
                        a page is searched
                When set to False:
                    -if just an agent_id and flow_name are specified
                        only the start page of the flow is searched
                    -if an agent_id, flow_name and page_name are specified
                        a page is searched
        Returns:
            - locator, dataframe of the results of where this string was found
        """

        if page_name:
            try:
                flows_map = self.flows.get_flows_map(
                    agent_id=agent_id, reverse=True
                )
            # check - maybe other error types here
            except ValueError:
                logging.error(
                    "%s is not a valid flow_name for agent %s",
                    flow_name,
                    agent_id,
                )
            try:
                pages_map = self.pages.get_pages_map(
                    flow_id=flows_map[flow_name], reverse=True
                )
                return self.search_conditionals_page(
                    page_id=pages_map[page_name], search=search
                )

            except ValueError:
                logging.error(
                    "%s is not a valid page_name for flow %s in agent %s",
                    page_name,
                    flow_name,
                    agent_id,
                )

        if flow_name:
            locator = pd.DataFrame()
            try:
                flows_map = self.flows.get_flows_map(
                    agent_id=agent_id, reverse=True
                )
                flow_search = self.search_conditionals_flow(
                    flow_id=flows_map[flow_name], search=search
                )
                flow_search.insert(0, "resource_name", flow_name)
                flow_search.insert(0, "resource_type", "flow")
                locator = locator.append(flow_search)
            except ValueError:
                logging.error(
                    "%s is not a valid flow_name for agent %s",
                    flow_name,
                    agent_id,
                )

            if flag_search_all:

                pages_map = self.pages.get_pages_map(
                    flow_id=flows_map[flow_name], reverse=True
                )
                for page in pages_map:
                    page_search = self.search_conditionals_page(
                        page_id=pages_map[page], search=search
                    )
                    time.sleep(0.5)
                    page_search.insert(0, "resource_name", page)
                    page_search.insert(0, "resource_type", "page")
                    locator = locator.append(page_search)

            return locator

        if flow_name is None and page_name is None and flag_search_all is True:
            locator = pd.DataFrame()

            flows_map = self.flows.get_flows_map(
                agent_id=agent_id, reverse=True
            )
            for flow in flows_map:
                flow_search = self.search_conditionals_flow(
                    flow_id=flows_map[flow], search=search
                )
                flow_search.insert(0, "resource_name", flow)
                flow_search.insert(0, "resource_type", "flow")
                locator = locator.append(flow_search)
                pages_map = self.pages.get_pages_map(
                    flow_id=flows_map[flow], reverse=True
                )
                for page in pages_map:
                    page_search = self.search_conditionals_page(
                        page_id=pages_map[page], search=search
                    )
                    time.sleep(0.5)
                    page_search.insert(0, "resource_name", page)
                    page_search.insert(0, "resource_type", "page")
                    locator = locator.append(page_search)
            return locator

        # not found
        return None


    def find_true_routes(self, agent_id: str = None):
        """This method extracts data to see if routes with no parameters have a
        true route or pages with parameters have a true route +
        page.params.status = "Final" route. Having these routes ensure a user
        can escape this page no matter what.

        Args:
          - agent_id: The properly formatted CX Agent ID

        Returns:
          - agent_results: dataframe with:
              flow_display_name: display name of the associated page
              page_display_name: display name of the page with the associated
                data
              webhook_entry_fulfillments: True if a page has a webhook on entry
                else False
              has_parameters: True if a page has parameters else False
              has_true_route: True if a page has a true route else False
              has_true_and_final_route: True if a page has a route with true
                + page.params.status=Final else False
        """

        if not agent_id:
            agent_id = self.agent_id

        agent_results = pd.DataFrame()
        flow_map = self.flows.get_flows_map(
                agent_id=agent_id, reverse=True
            )

        for flow_display_name in flow_map.keys():
            flow_scan = self._find_true_routes_flow_level(
                flow_display_name, flow_map
            )
            agent_results = agent_results.append(flow_scan)
        return agent_results


    # Event handlers Main Function
    def find_event_handlers(self):
        """This method extracts event handlers at the flow, page and parameter
        level and displays data about their associated event. A user can use
        this data to spot patterns in event types and look for detrimental
        patterns.

        Args:
          - agent_id must specify agent id when instantiating the class

        Returns:
          - dictionary with flow, page and parameter events
        """
        event_handler_scan = {
            "flow": self._flow_level_handlers(),
            "page": self._page_level_handlers(),
            "parameter": self._parameter_level_handlers(),
        }

        return event_handler_scan



"""Test Class for Agent Functions in SCRAPI lib."""

# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from datetime import datetime
# from typing import List, Dict

import pytest
import google.cloud.dialogflowcx_v3beta1.types as types

DEV = True  # Set flag to disable some tests while in development

today_time = datetime.now().strftime("%d%m%Y_%H%M%S")
AGENT_NAME = "DFCX SCRAPI - TEMP TEST AGENT {}".format(today_time)
CREDS_PATH = None
PROJECT_ID = None

pytest.temp_agent = None

def test_create_agent(agents, project):
    """Tests the SCRAPI method create_agent from core/agents.py"""
    # agents = Agents(creds_path=creds)
    pytest.temp_agent = agents.create_agent(
        project, AGENT_NAME
    )

    assert isinstance(pytest.temp_agent, types.Agent)
    assert pytest.temp_agent.display_name == AGENT_NAME

    return pytest.temp_agent


def test_restore_agent(agents, ops, gcs_bucket):
    """Tests the SCRAPI method restore_agent from core/agents.py"""
    lro = agents.restore_agent(
        pytest.temp_agent.name,
        gcs_bucket_uri="gs://{}/sample_agent.blob".format(gcs_bucket)
    )

    time.sleep(4)
    res = ops.get_lro(lro)
    print(res)

    assert res["done"]


def test_get_agent(agents):
    """Tests the SCRAPI method get_agent from core/agents.py"""
    agent = agents.get_agent(pytest.temp_agent.name)

    assert isinstance(agent, types.Agent)
    assert pytest.temp_agent.display_name == AGENT_NAME

def test_get_agent_by_display_name(agents, project):
    """Test that we can retrieve agent by display name"""
    region = pytest.temp_agent.name.split("/")[3]
    agent = agents.get_agent_by_display_name(
        project_id=project,
        display_name=pytest.temp_agent.display_name,
        region=region
    )

    assert isinstance(agent, types.Agent)
    assert agent.display_name == AGENT_NAME

def test_get_agent_by_display_name_dupe(agents, project, caplog):
    """Test this function when providing a display name that is ambiguous and
    exists in multiple regions"""
    with caplog.at_level(logging.WARNING):
        agents.get_agent_by_display_name(
            project_id=project,
            display_name="SCRAPI - CI Duplicate")

    assert "Found multiple agents with the display name" in caplog.text

def test_get_agent_by_display_name_approx(agents, project, caplog):
    """test this function with approximate matching of agent name."""
    with caplog.at_level(logging.WARNING):
        agents.get_agent_by_display_name(
            project_id=project,
            display_name="sCRAPI - CI Duplicate"
        )

    assert "display_name is case-sensitive" in caplog.text

def test_export_agent(agents, ops, gcs_bucket):
    """Tests the SCRAPI method export_agent from core/agents.py"""
    lro = agents.export_agent(
        pytest.temp_agent.name,
        "gs://{}/testing_export.json".format(gcs_bucket),
    )

    time.sleep(4)

    assert ops.get_lro(lro)["done"]


def test_update_agent(agents):
    """Tests the SCRAPI method update_agent from core/agents.py"""
    temp_name = AGENT_NAME + "_UPDATED"
    res = agents.update_agent(
        pytest.temp_agent.name, display_name=temp_name
    )

    assert res.display_name == temp_name


def test_delete_agent(agents):
    """Tests the SCRAPI method delete_agent from core/agents.py"""
    res = agents.delete_agent(pytest.temp_agent.name)

    print(res)

    assert pytest.temp_agent.name in res

def test_list_agents(agents, project):
    """Tests the SCRAPI method list_agents from core/agents.py"""
    location_id = "projects/{}/locations/global".format(project)
    agents = agents.list_agents(location_id)

    assert isinstance(agents[0], types.agent.Agent)
